from .diapyr import DI, types
from .iface import MissingAnnotationException, UnsatisfiableRequestException

DI = DI
'Convenience import.'
assert MissingAnnotationException
types = types
'Convenience import.'
assert UnsatisfiableRequestException
